import numpy
import math
import os
import sys
import subprocess
from sets import Set
from numpy import *

####################################################################	
def loadOff(fileName):
	file = open(fileName,"r")	
	tag = 0
	header = 0
	numVert = 0
	numTri = 0
	hasColors = 0
	print "Loading mesh..."
	
	vert = numpy.ones(1)
	tri = numpy.ones(1)
	
	while 1:
		line = file.readline()		
		if not line:
			break
			
		if (len(line)<=2):
			continue
							
		if (tag==0 and line[0:3]=="OFF"):
			tag = 1
			continue		
			
		if (tag==0 and line[0:4]=="COFF"):
			tag = 1
			hasColors = 1
			continue			

		if (line[0]=='#'):
			print "Mesh comment line: ", line
			continue
		
		# if OFF is read then the header can be read
		if (header == 0 and tag == 1):		
			v = line.split()
									
			numVert = int(v[0])
			numTri = int(v[1])
			
			if (numVert <=0 or numTri<=0):			
				print "Number of triangles or vertices <=0 !"
				return 			
			else:
				vert = numpy.ones((numVert,3),dtype=float)
				tri = numpy.ones((numTri,3),dtype=int)
				vertexTrianglesList = [0]*numVert				
				for i in range(numVert):
					vertexTrianglesList[i]=Set()
			header = 1;
			break;

	if (header !=1 or tag !=1):
		print  "Cannot read OFF header or number of vertices/triangles, stop reading"
		return
	
	colors = []
	
	if (hasColors):
		colors = numpy.zeros((numVert,3))
			
	# read vertices
	for i in range(numVert):
		line = file.readline()
		if (len(line)<=1):		
			i = i - 1
			continue
		v = line.split()
		
		vert[i,0] = float(v[0])
		vert[i,1] = float(v[1])
		vert[i,2] = float(v[2])		
	
		if (hasColors):
			colors[i,0] = float(v[3])/255
			colors[i,1] = float(v[4])/255
			colors[i,2] = float(v[5])/255
			#print colors[i,:]
		
	# read triangles
	for i in range(numTri):
		line = file.readline()
		if (len(line)<=1):		
			i = i - 1
			continue

		v = line.split()
		tri[i,0] = int(v[1])
		tri[i,1] = int(v[2])
		tri[i,2] = int(v[3])
		
		vertexTrianglesList[tri[i,0]].add(i)
		vertexTrianglesList[tri[i,1]].add(i)
		vertexTrianglesList[tri[i,2]].add(i)

		if (int(v[0])!=3):		
			print "Non triangular mesh, stop reading"
			return
				
	file.close()
	
	planeNormals = numpy.zeros((numTri,3))
	
	# compute plane normals from triangles 
	for it in range(numTri):
		p1 = vert[tri[it][0],:]
		p2 = vert[tri[it][1],:]
		p3 = vert[tri[it][2],:]

		# plane three points
		planeNormals[it,0] = p1[1]*(p2[2] - p3[2]) + p2[1]*(p3[2] - p1[2]) + p3[1]*(p1[2] - p2[2])
		planeNormals[it,1] = p1[2]*(p2[0] - p3[0]) + p2[2]*(p3[0] - p1[0]) + p3[2]*(p1[0] - p2[0])
		planeNormals[it,2] = p1[0]*(p2[1] - p3[1]) + p2[0]*(p3[1] - p1[1]) + p3[0]*(p1[1] - p2[1])
	
	# now estimates vertex normals
	vertNormals = numpy.zeros((numVert,3))
	meanNormal = numpy.zeros(3)
	
	for iv in range(numVert):		
		p = vert[iv,:];						
		meanNormal[0]=0;
		meanNormal[1]=0;
		meanNormal[2]=0;
		
		for triangleID in vertexTrianglesList[iv]:				
				meanNormal = meanNormal+planeNormals[triangleID,:]

		vlen = len(vertexTrianglesList[iv])
		vertNormals[iv,0]=meanNormal[0]/vlen;
		vertNormals[iv,1]=meanNormal[1]/vlen;
		vertNormals[iv,2]=meanNormal[2]/vlen;

		v = vertNormals[iv,:]
		
		# normalization
		lnorm = sqrt(dot(v,v))
		vertNormals[iv,0]=vertNormals[iv,0]/lnorm;
		vertNormals[iv,1]=vertNormals[iv,1]/lnorm;
		vertNormals[iv,2]=vertNormals[iv,2]/lnorm;		

	return numVert,numTri,tri,vert,-vertNormals,hasColors,colors

####################################################################
def loadPhiMap(fileName):
	file = open(fileName,"r")	
	print "Loading phimap..."
	
	ind = -1
	barx = 0
	bary = 0
	barz = 0
	grid = 0
	scale = 0
	
	phi = numpy.ones((1,1,1))
	
	while 1:
		line = file.readline()		
		if not line:
			break
			
		if (ind==-1):
			ind = 0
			v = line.split()
			scale = float(v[0])
			barx = float(v[1])
			bary = float(v[2])
			barz = float(v[3])
			grid = int(v[4])
			print scale,barx,bary,barz,grid
			phi = numpy.ones((grid,grid,grid))
			break
			
	for i in range(grid):
		for j in range(grid):
			for k in range(grid):
				line = file.readline()		
				v = line.split()
				ff = float(v[3])
				phi[i,j,k] = ff

	offset = (grid-1)/(2*scale)
	x0 = barx-offset;
	y0 = bary-offset;
	z0 = barz-offset;
	
	xn = barx+offset;
	yn = bary+offset;
	zn = barz+offset;
		
	x = linspace(x0,xn,grid);
	y = linspace(y0,yn,grid);
	z = linspace(z0,zn,grid);
			
	xg,yg,zg = numpy.mgrid[x0:xn:complex(grid),y0:yn:complex(grid),z0:zn:complex(grid)];
		
	file.close()
	
	return x,y,z,xg,yg,zg,x0,y0,z0,1/scale,phi,grid		

######################################################################
	
# transform off to idtf 
# mode 1 only off, mode 2 off+phi map, mode 3 coloured off

def main():
	
	if (len(sys.argv)<4):
		print "\n"
		print "Transform a .off file or .off+phi map into an idtf file \ntogether with a corresponding latex source"
		print "produced files are surf.idtf and surf.tex. Then convert into u3d and run pdflatex"
		print "use latex with movie15 package to build an interactive pdf."
		print "\n Usage:"
		print "python off2pdf.py meshFileName.off mode ./path_to_converter/converter"
		print "- mode 1 is only off, mode 2 is off+DelPhi ASCII phi map, mode 3 load .off with vertex color"
		print "In VMD you can save in obj format, then to convert an obj file rendered by VMD to a coloured off use meshlabserver.exe -i vmdscene.obj -o vmdscene.off -om vc"
		print "Example: python off2pdf.py triangulatedSurf.off 2 ./bin/w32/IDTFConverter.exe"
		print "\n"
		return
		
	meshFileName = sys.argv[1]
	mode = int(sys.argv[2])
	converter = sys.argv[3]
	phiFileName = 'fort.14'
	numVert,numTri,tri,vert,vertNormals,hasColors,palette = loadOff(meshFileName)

	coo = numpy.zeros((3,1))
	v = numpy.zeros((3,1))
	coo[0]=numpy.mean(vert[:,0])
	coo[1]=numpy.mean(vert[:,1])
	coo[2]=numpy.mean(vert[:,2])

	v[0] = max(abs(vert[:,0]))
	v[1] = max(abs(vert[:,1]))
	v[2] = max(abs(vert[:,2]))

	vmax = max(vert[:,2])
	vmin = min(vert[:,2])

	# camera position
	c2c = (coo-v)

	dist = sqrt(dot(c2c[0,:],c2c[0,:]))*5
	colorIndex = numpy.zeros((numTri,3))
			
	numNorm = numTri*3

	x = []
	y = []
	z = []
	xg = []
	yg = []
	zg = []
	xmin = 0
	ymin = 0
	zmin = 0

	xmax = 0
	ymax = 0
	zmax = 0
	side = 0
	grid = 0
	phiMap = []

	bg = numpy.zeros(3)

	if (mode==2):
		
		x,y,z,xg,yg,zg,xmin,ymin,zmin,side,phi,grid = loadPhiMap(phiFileName)
		phiMap = numpy.ones(numVert)
		
		# for each vertex get the proper phi value
		for i in range(numVert):
			ix = (vert[i,0]-xmin)/side
			iy = (vert[i,1]-ymin)/side
			iz = (vert[i,2]-zmin)/side
			phiMap[i]=phi[ix,iy,iz]		
			
			# ix = floor((vert[i,0]-xmin)/side)
			# iy = floor((vert[i,1]-ymin)/side)
			# iz = floor((vert[i,2]-zmin)/side)	
		
			# p010 = phi[ix,iy,iz]		
			# p110 = phi[ix+1,iy,iz]		
			# p100 = phi[ix+1,iy+1,iz]		
			# p000 = phi[ix,iy+1,iz]		
		
			# p011 = phi[ix,iy,iz+1]		
			# p111 = phi[ix+1,iy,iz+1]		
			# p101 = phi[ix+1,iy+1,iz+1]		
			# p001 = phi[ix,iy+1,iz+1]		
				
			# dx = (vert[i,0]-x[ix])/side;
			# dy = (vert[i,1]-y[iy+1])/side;
			# dz = (vert[i,2]-z[iz])/side;
	
			# c0 = p000;
			# c1 = p100-p000;
			# c2 = p010-p000;
			# c3 = p001-p000;
			# c4 = p110-p010-p100+p000;
			# c5 = p011-p001-p010+p000;
			# c6 = p101-p001-p100+p000;
			# c7 = p111-p011-p101-p110+p100+p001+p010-p000;

			# val=c0+c1*dx+c2*dy+c3*dz+c4*dx*dy+c5*dy*dz+c6*dz*dx+c7*dx*dy*dz;		
			# phiMap[i]=val
		
		minV = min(phiMap)
		maxV = max(phiMap)
		maxV = max(abs(minV),abs(maxV))
		minV = -maxV
		
		print minV,maxV
		print grid,len(xg),len(yg),len(zg),len(phi[:,1,1]),len(phi[1,1,:]),len(phi[1,:,1])

	if (mode==1):
		numCol = 0
		bg[0]=1
		bg[1]=1
		bg[2]=1
		
	elif(mode==2):
		numNorm=0
		numCol = 256	
		palette = numpy.ones((numCol,3))	
		# potential palette
		# the less value the more the contrast
		threshold = 10	
		t1 = 100
		t2 = 128-t1
		v1 = numpy.concatenate((numpy.linspace(0,threshold,t1),numpy.linspace(threshold,255,t2),255*numpy.ones(128)))		
		#Electrostatic Potential blue-white-red Colormap		
		palette[:,0] = v1[::-1]/255
		v2 = numpy.concatenate((numpy.linspace(0,threshold,t1),numpy.linspace(threshold,255,t2),numpy.linspace(255,threshold,t2),numpy.linspace(threshold,0,t1)))	
		palette[:,1] = v2[::-1]/255
		v3 = numpy.concatenate((255*numpy.ones(128),numpy.linspace(255,threshold,t2),numpy.linspace(threshold,0,t1)))			
		palette[:,2] = v3[::-1]/255

		# get color index for each vertex and for each triangle
		for i in range(numTri):	
			ind =(phiMap[tri[i,0]]-minV)/(maxV-minV)*(numCol-1)
			colorIndex[i,0]=int(ind)
			ind =(phiMap[tri[i,1]]-minV)/(maxV-minV)*(numCol-1)
			colorIndex[i,1]=int(ind)
			ind =(phiMap[tri[i,2]]-minV)/(maxV-minV)*(numCol-1)
			colorIndex[i,2]=int(ind)
			
	elif(mode==3):
		# get unique colors map
		numNorm = 0
		numCol = numVert
		index = 0
		for i in range(numTri):			
			colorIndex[i,0]=tri[i,0]			
			colorIndex[i,1]=tri[i,1]			
			colorIndex[i,2]=tri[i,2]

				
	#print "Num Colors ",numCol

	#print palette
	
	f = open('surf.idtf', 'w')
	f.write('FILE_FORMAT "IDTF"\n' \
	'FORMAT_VERSION 100\n' \
	'NODE "MODEL" {\n' \
	'	NODE_NAME "Mesh"\n' \
	'	PARENT_LIST {\n' \
	'		PARENT_COUNT 1\n' \
	'		PARENT 0 {\n' \
	'			PARENT_NAME "<NULL>"\n' \
	'			PARENT_TM {\n' \
	'				1.000000 0.000000 0.000000 0.000000\n' \
	'				0.000000 1.000000 0.000000 0.000000\n' \
	'				0.000000 0.000000 1.000000 0.000000\n' \
	'				0.000000 0.000000 0.000000 1.000000\n' \
	'			}\n' \
	'		}\n' \
	'	}\n' \
	'	RESOURCE_NAME "MyMesh"\n' \
	'}\n' \
	'\n' \
	'RESOURCE_LIST "MODEL" {\n'\
	'	RESOURCE_COUNT 1\n'\
	'	RESOURCE 0 {\n'\
	'		RESOURCE_NAME "MyMesh"\n'\
	'		MODEL_TYPE "MESH"\n'\
	'		MESH {\n'\
	'			FACE_COUNT %d\n'\
	'			MODEL_POSITION_COUNT %d\n'\
	'			MODEL_NORMAL_COUNT %d\n'\
	'			MODEL_DIFFUSE_COLOR_COUNT %d\n'\
	'			MODEL_SPECULAR_COLOR_COUNT 0\n'\
	'			MODEL_TEXTURE_COORD_COUNT 0\n'\
	'			MODEL_BONE_COUNT 0\n'\
	'			MODEL_SHADING_COUNT 1\n'\
	'			MODEL_SHADING_DESCRIPTION_LIST {\n'\
	'				SHADING_DESCRIPTION 0 {\n'\
	'					TEXTURE_LAYER_COUNT 0\n'\
	'					SHADER_ID 1\n'\
	'				}\n'\
	'			}\n' %(numTri,numVert,numNorm,numCol))

	f.write('			MESH_FACE_POSITION_LIST {\n')
	for i in range(numTri):
		f.write('					%d %d %d\n'%(tri[i,0],tri[i,1],tri[i,2]))
	f.write('				}\n')

	f.write('')
	if (numNorm>0):
		f.write('			MESH_FACE_NORMAL_LIST {\n')
		for i in range(numTri):
			f.write('				%d %d %d\n'%(i*3,i*3+1,i*3+2))
		f.write('			}\n')

	f.write('				MESH_FACE_SHADING_LIST {\n')
	for i in range(numTri):
		f.write('				0\n')
	f.write('			}\n')

	f.write('')
	if (numCol>0):
		f.write('				MESH_FACE_DIFFUSE_COLOR_LIST {\n')
		for i in range(numTri):
			f.write('			%d %d %d\n'%(colorIndex[i,0],colorIndex[i,1],colorIndex[i,2]))
		f.write('			}\n')
		
	f.write('				MODEL_POSITION_LIST {\n')
	for i in range(numVert):
		f.write('				%f %f %f\n'%(vert[i,0],vert[i,1],vert[i,2]))
	f.write('			}\n')

	f.write('')
	if (numCol>0):
		f.write('				MODEL_DIFFUSE_COLOR_LIST {\n')
		for i in range(numCol):
			f.write('				%f %f %f\n'%(palette[i,0],palette[i,1],palette[i,2]))
		f.write('			}\n')

	f.write('')
	if (numNorm>0):
		f.write('				MODEL_NORMAL_LIST {\n')
		for i in range(numTri):
			f.write('				%f %f %f\n'%(vertNormals[tri[i,0],0],vertNormals[tri[i,0],1],vertNormals[tri[i,0],2]))
			f.write('				%f %f %f\n'%(vertNormals[tri[i,1],0],vertNormals[tri[i,1],1],vertNormals[tri[i,1],2]))
			f.write('				%f %f %f\n'%(vertNormals[tri[i,2],0],vertNormals[tri[i,2],1],vertNormals[tri[i,2],2]))
		f.write('			}\n')

	f.write( '		}\n'\
	'	}\n'\
	'}\n'\
	'\n')

	f.write( 'RESOURCE_LIST "SHADER" {\n'\
	'	RESOURCE_COUNT 1\n'\
	'	RESOURCE 0 {\n'\
	'		RESOURCE_NAME "Box010"\n'\
	'		ATTRIBUTE_USE_VERTEX_COLOR "TRUE"\n'\
	'		SHADER_MATERIAL_NAME "Box010"\n'\
	'		SHADER_ACTIVE_TEXTURE_COUNT 0\n'\
	'	}\n'\
	'}\n'\
	'\n'\
	'RESOURCE_LIST "MATERIAL" {\n'\
	'	RESOURCE_COUNT 1\n'\
	'	RESOURCE 0 {\n'\
	'		RESOURCE_NAME "Box010"\n'\
	'		MATERIAL_AMBIENT 0.0000000 0.0000000 0.0000000\n'\
	'		MATERIAL_DIFFUSE 0.999 0.999 0.999\n'\
	'		MATERIAL_SPECULAR 0.0000000 0.0000000 0.0000000\n'\
	'		MATERIAL_EMISSIVE 0.000000 0.000000 0.000000\n'\
	'		MATERIAL_REFLECTIVITY 0.000000\n'\
	'		MATERIAL_OPACITY 1.000000\n'\
	'	}\n'\
	'}\n'\
	'\n'\
	'MODIFIER "SHADING" {\n'\
	'	MODIFIER_NAME "Mesh"\n'\
	'	PARAMETERS {\n'\
	'		SHADER_LIST_COUNT 1\n'\
	'		SHADER_LIST_LIST {\n'\
	'			SHADER_LIST 0 {\n'\
	'				SHADER_COUNT 1\n'\
	'				SHADER_NAME_LIST {\n'\
	'					SHADER 0 NAME: "Box010"\n'\
	'				}\n'\
	'			}\n'\
	'		}\n'\
	'	}\n'\
	'}\n' )
		

	f.close()

	f = open('surf.tex','w')
	str1 = '\documentclass[a4paper]{article}\n'\
	'\usepackage[3D]{movie15}\n'\
	'\usepackage[UKenglish]{babel}\n'\
	'\usepackage[colorlinks=true]{hyperref}\n'\
	'\\begin{document}\n'\
	'\\begin{center}\n'\
	'\includemovie[poster,toolbar,label=surf.u3d,text=(surf.u3d),3Droo=%f,3Dc2c=%f %f %f,3Dcoo=%f %f %f,3Dbg=%f %f %f,3Drender=Solid,3Dlights=White]{\linewidth}\n'\
	'{\linewidth}{surf.u3d}\n'\
	'\end{center}\n'\
	'\end{document}'%(dist,c2c[0],c2c[1],c2c[2],coo[0],coo[1],coo[2],bg[0],bg[1],bg[2])
	f.write(str1)
	f.close()

	# now convert from idtf to u3d
	execString = '"%s" -input "%s/surf.idtf" -output "%s/surf.u3d"'%(converter,os.getcwd(),os.getcwd())
	print execString
	subprocess.call(execString,shell=True)	
	execString = 'pdflatex -quiet surf.tex'
	print execString
	# run latex
	subprocess.call(execString,shell=True)	

if __name__ == '__main__':
	main()